/*
 * Simple synchronous userspace interface to SPI devices
 *
 * Copyright (C) 2006 SWAPP
 *	Andrea Paterniani <a.paterniani@swapp-eng.it>
 * Copyright (C) 2007 David Brownell (simplification, cleanup)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/mutex.h>
#include <linux/slab.h>
#include <linux/compat.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/gpio.h>
#include <linux/spi/spi.h>
#include <linux/spi/spidev.h>
#include <linux/delay.h>
#include <linux/uaccess.h>


/*
 * This supports access to SPI devices using normal userspace I/O calls.
 * Note that while traditional UNIX/POSIX I/O semantics are half duplex,
 * and often mask message boundaries, full SPI support requires full duplex
 * transfers.  There are several kinds of internal message boundaries to
 * handle chipselect management and other protocol options.
 *
 * SPI has a character major number assigned.  We allocate minor numbers
 * dynamically using a bitmask.  You must use hotplug tools, such as udev
 * (or mdev with busybox) to create and destroy the /dev/spidevB.C device
 * nodes, since there is no fixed association of minor numbers with any
 * particular SPI bus or device.
 */
 
#define lcd_ili9341
#ifdef lcd_st7789
#define GPIO_RST				59
#define GPIO_RS					38
#define LCD_W					240
#define LCD_H					320
#elif defined	lcd_ili9341
#define GPIO_RST				42
#define GPIO_RS					11
#define LCD_W					160
#define LCD_H					128
#elif defined lcd_ili9163
#define GPIO_RST		24
#define GPIO_RS			42
#define GPIO_BL 				11
#define LCD_W					160
#define LCD_H					128
#endif

#define SPIDEV_MAJOR			153	/* assigned */
#define N_SPI_MINORS			32	/* ... up to 256 */

#define FRAMEBUFF_SIZE 128*21

#define SetColumnAddr		0x04
#define SetTempVopCtrl		0x10
#define SetTempFRCtrl		0x13
#define SetSensorCtrl		0x15
#define SetTempCompensation	0x20
#define SetPumpCtrl			0x2D
#define SetScrollLine		0x40
#define SetPageAddr			0x60
#define SetPMCtrl			0x81
#define SetCOMScan			0x86
#define SetRAMAddrCtrl		0x8B
#define SetLCDMapCtrl		0xc0
#define SetLCDBiasRatio		0xED
#define SetCOMEnd			0xF1
#define SetNlineInversion	0xC8
#define SetDisplayEnable	0xC9
#define SetAllPixelON		0xA5
#define SetAllPixelOFF		0xA4
	
static DECLARE_BITMAP(minors, N_SPI_MINORS);


/* Bit masks for spi_device.mode management.  Note that incorrect
 * settings for some settings can cause *lots* of trouble for other
 * devices on a shared bus:
 *
 *  - CS_HIGH ... this device will be active when it shouldn't be
 *  - 3WIRE ... when active, it won't behave as it should
 *  - NO_CS ... there will be no explicit message boundaries; this
 *	is completely incompatible with the shared bus model
 *  - READY ... transfers may proceed when they shouldn't.
 *
 * REVISIT should changing those flags be privileged?
 */
#define SPI_MODE_MASK		(SPI_CPHA | SPI_CPOL | SPI_CS_HIGH \
				| SPI_LSB_FIRST | SPI_3WIRE | SPI_LOOP \
				| SPI_NO_CS | SPI_READY | SPI_TX_DUAL \
				| SPI_TX_QUAD | SPI_RX_DUAL | SPI_RX_QUAD)

struct spidev_data {
	dev_t			devt;
	spinlock_t		spi_lock;
	struct spi_device	*spi;
	struct list_head	device_entry;

	/* TX/RX buffers are NULL unless this device is open (users > 0) */
	struct mutex		buf_lock;
	unsigned		users;
	u8			*tx_buffer;
	u8			*rx_buffer;
};

typedef struct _spi_lcd_cmd{
	unsigned char  reg_addr;  
	unsigned char len;
 	int delay_ms; 
}spi_lcd_cmd;

static LIST_HEAD(device_list);
static DEFINE_MUTEX(device_list_lock);

static unsigned bufsiz = 4096;
module_param(bufsiz, uint, S_IRUGO);
MODULE_PARM_DESC(bufsiz, "data bytes in biggest supported SPI message");

/*
 * This can be used for testing the controller, given the busnum and the
 * cs required to use. If those parameters are used, spidev is
 * dynamically added as device on the busnum, and messages can be sent
 * via this interface.
 */
static int busnum = 6;
module_param(busnum, int, S_IRUGO);
MODULE_PARM_DESC(busnum, "bus num of the controller");

static int chipselect = 0;
module_param(chipselect, int, S_IRUGO);
MODULE_PARM_DESC(chipselect, "chip select of the desired device");

static int maxspeed = 4800000;
module_param(maxspeed, int, S_IRUGO);
MODULE_PARM_DESC(maxspeed, "max_speed of the desired device");

static int spimode = 0;
module_param(spimode, int, S_IRUGO);
MODULE_PARM_DESC(spimode, "mode of the desired device");

static struct spi_device *spi;

/*-------------------------------------------------------------------------*/

/*
 * We can't use the standard synchronous wrappers for file I/O; we
 * need to protect against async removal of the underlying spi_device.
 */
static void spidev_complete(void *arg)
{
	complete(arg);
}

static ssize_t
spidev_sync(struct spidev_data *spidev, struct spi_message *message)
{
	DECLARE_COMPLETION_ONSTACK(done);
	int status;
	
	message->complete = spidev_complete;
	message->context = &done;
	spin_lock_irq(&spidev->spi_lock);
	if (spidev->spi == NULL)
		status = -ESHUTDOWN;
	else
	{
		status = spi_async(spidev->spi, message);
	}
	spin_unlock_irq(&spidev->spi_lock);
	if (status == 0) {
		wait_for_completion(&done);
		status = message->status;
		if (status == 0)
			status = message->actual_length;
	}	
	return status;
}

static inline ssize_t
spidev_sync_write(struct spidev_data *spidev,void *data, size_t len)
{
	struct spi_transfer	t = {
			.tx_buf		= data,
			.len		= len,
		};
	struct spi_message	m;

	spi_message_init(&m);
	spi_message_add_tail(&t, &m);
	return spidev_sync(spidev, &m);
}

static inline ssize_t
spidev_sync_read(struct spidev_data *spidev, size_t len)
{
	struct spi_transfer	t = {
			.rx_buf		= spidev->rx_buffer,
			.len		= len,
		};
	struct spi_message	m;

	spi_message_init(&m);
	spi_message_add_tail(&t, &m);
	return spidev_sync(spidev, &m);
}

/*-------------------------------------------------------------------------*/

/* Read-only message with current device setup */
static ssize_t
spidev_lcd_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
	struct spidev_data	*spidev;
	ssize_t			status = 0;

	/* chipselect only toggles at start or end of operation */
	if (count > FRAMEBUFF_SIZE)
		return -EMSGSIZE;

	spidev = filp->private_data;

	mutex_lock(&spidev->buf_lock);
	status = FRAMEBUFF_SIZE;
	if (status > 0) {
		unsigned long	missing;

		missing = copy_to_user(buf, spidev->tx_buffer, status);
		if (missing == status)
			status = -EFAULT;
		else
			status = status - missing;
	}
	mutex_unlock(&spidev->buf_lock);

	return status;
}

/* Write-only message with current device setup */
int write_data_buffer(struct spi_device *spi,unsigned char *data ,int len);
void write_image(struct spi_device *spi,unsigned char data );
void write_command(struct spi_device *spi, unsigned char cmd);
void write_data(struct spi_device *spi,unsigned char data );

static ssize_t
spidev_lcd_write(struct file *filp, const char __user *buf,
		size_t count, loff_t *f_pos)
{
	struct spidev_data	*spidev;
	ssize_t			status = 0;
	unsigned long		missing;
	static int TotalCount= 0;

	/* chipselect only toggles at start or end of operation */
	if (count > FRAMEBUFF_SIZE)
		return -EMSGSIZE;

	spidev = filp->private_data;

	mutex_lock(&spidev->buf_lock);
	missing = copy_from_user(spidev->tx_buffer, buf, count);
	if (missing == 0)
	{
		if(TotalCount == 0) 
		{
			write_command(spidev->spi, SetColumnAddr);
			write_data(spidev->spi, 0x00);
			write_command(spidev->spi, SetPageAddr);	
			write_data(spidev->spi, 0x00);	

			write_command(spidev->spi, 0x01);	//JHL-20200623
		}
		status =write_data_buffer(spidev->spi,spidev->tx_buffer,count); 
	}
	else
		status = -EFAULT;
	mutex_unlock(&spidev->buf_lock);
	TotalCount= TotalCount + count;

	if(TotalCount == 2688) TotalCount= 0;

	return status;
}

static int spidev_message(struct spidev_data *spidev,
		struct spi_ioc_transfer *u_xfers, unsigned n_xfers)
{
	struct spi_message	msg;
	struct spi_transfer	*k_xfers;
	struct spi_transfer	*k_tmp;
	struct spi_ioc_transfer *u_tmp;
	unsigned		n, total;
	u8			*tx_buf, *rx_buf;
	int			status = -EFAULT;

	spi_message_init(&msg);
	k_xfers = kcalloc(n_xfers, sizeof(*k_tmp), GFP_KERNEL);
	if (k_xfers == NULL)
		return -ENOMEM;

	/* Construct spi_message, copying any tx data to bounce buffer.
	 * We walk the array of user-provided transfers, using each one
	 * to initialize a kernel version of the same transfer.
	 */
	tx_buf = spidev->tx_buffer;
	rx_buf = spidev->rx_buffer;
	total = 0;
	for (n = n_xfers, k_tmp = k_xfers, u_tmp = u_xfers;
			n;
			n--, k_tmp++, u_tmp++) {
		k_tmp->len = u_tmp->len;

		total += k_tmp->len;
		/* Check total length of transfers.  Also check each
		 * transfer length to avoid arithmetic overflow.
		 */
		if (total > bufsiz || k_tmp->len > bufsiz) {
			status = -EMSGSIZE;
			goto done;
		}

		if (u_tmp->rx_buf) {
			k_tmp->rx_buf = rx_buf;
			if (!access_ok(VERIFY_WRITE, (u8 __user *)
						(uintptr_t) u_tmp->rx_buf,
						u_tmp->len))
				goto done;
		}
		if (u_tmp->tx_buf) {
			k_tmp->tx_buf = tx_buf;
			if (copy_from_user(tx_buf, (const u8 __user *)
						(uintptr_t) u_tmp->tx_buf,
					u_tmp->len))
				goto done;
		}
		tx_buf += k_tmp->len;
		rx_buf += k_tmp->len;

		k_tmp->cs_change = !!u_tmp->cs_change;
		k_tmp->tx_nbits = u_tmp->tx_nbits;
		k_tmp->rx_nbits = u_tmp->rx_nbits;
		k_tmp->bits_per_word = u_tmp->bits_per_word;
		k_tmp->delay_usecs = u_tmp->delay_usecs;
		k_tmp->speed_hz = u_tmp->speed_hz;
#ifdef VERBOSE
		dev_dbg(&spidev->spi->dev,
			"  xfer len %zd %s%s%s%dbits %u usec %uHz b > spi_message_add_tail()\n",
			u_tmp->len,
			u_tmp->rx_buf ? "rx " : "",
			u_tmp->tx_buf ? "tx " : "",
			u_tmp->cs_change ? "cs " : "",
			u_tmp->bits_per_word ? : spidev->spi->bits_per_word,
			u_tmp->delay_usecs,
			u_tmp->speed_hz ? : spidev->spi->max_speed_hz);
#endif
		spi_message_add_tail(k_tmp, &msg);
	}

	status = spidev_sync(spidev, &msg);
	if (status < 0)
		goto done;

	/* copy any rx data out of bounce buffer */
	rx_buf = spidev->rx_buffer;
	for (n = n_xfers, u_tmp = u_xfers; n; n--, u_tmp++) {
		if (u_tmp->rx_buf) {
			if (__copy_to_user((u8 __user *)
					(uintptr_t) u_tmp->rx_buf, rx_buf,
					u_tmp->len)) {
				status = -EFAULT;
				goto done;
			}
		}
		rx_buf += u_tmp->len;
	}
	status = total;

done:
	kfree(k_xfers);
	return status;
}

static long
spidev_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int			err = 0;
	int			retval = 0;
	struct spidev_data	*spidev;
	struct spi_device	*spi;
	u32			tmp;
	unsigned		n_ioc;
	struct spi_ioc_transfer	*ioc;
	int		ContrastValue;

	printk("[spidev_ioctl] cmd(%4x), _IOC_TYPE(%d),_IOC_DIR(%d), _IOC_SIZE(%d)\n", cmd, _IOC_TYPE(cmd), _IOC_DIR(cmd), _IOC_SIZE(cmd));
	
	/* Check type and command number */
	if (_IOC_TYPE(cmd) != SPI_IOC_MAGIC)
		return -ENOTTY;

	/* Check access direction once here; don't repeat below.
	 * IOC_DIR is from the user perspective, while access_ok is
	 * from the kernel perspective; so they look reversed.
	 */
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE,
				(void __user *)arg, _IOC_SIZE(cmd));
	if (err == 0 && _IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ,
				(void __user *)arg, _IOC_SIZE(cmd));				
	if (err)
		return -EFAULT;
	/* guard against device removal before, or while,
	 * we issue this ioctl.
	 */
	spidev = filp->private_data;
	spin_lock_irq(&spidev->spi_lock);
	spi = spi_dev_get(spidev->spi);
	spin_unlock_irq(&spidev->spi_lock);

	if (spi == NULL)
		return -ESHUTDOWN;

	/* use the buffer lock here for triple duty:
	 *  - prevent I/O (from us) so calling spi_setup() is safe;
	 *  - prevent concurrent SPI_IOC_WR_* from morphing
	 *    data fields while SPI_IOC_RD_* reads them;
	 *  - SPI_IOC_MESSAGE needs the buffer locked "normally".
	 */
	mutex_lock(&spidev->buf_lock);
	
	switch (cmd) {
	/* read requests */
	case SPI_IOC_RD_MODE:
		retval = __put_user(spi->mode & SPI_MODE_MASK,
					(__u8 __user *)arg);
		break;
	case SPI_IOC_RD_MODE32:	
		retval = __put_user(spi->mode & SPI_MODE_MASK,
					(__u32 __user *)arg);
		break;
	case SPI_IOC_RD_LSB_FIRST:	
		retval = __put_user((spi->mode & SPI_LSB_FIRST) ?  1 : 0,
					(__u8 __user *)arg);
		break;
	case SPI_IOC_RD_BITS_PER_WORD:	
		retval = __put_user(spi->bits_per_word, (__u8 __user *)arg);
		break;
	case SPI_IOC_RD_MAX_SPEED_HZ:	
		retval = __put_user(spi->max_speed_hz, (__u32 __user *)arg);
		break;

	/* write requests */
	case SPI_IOC_WR_MODE:
	case SPI_IOC_WR_MODE32:	
		if (cmd == SPI_IOC_WR_MODE)
			retval = __get_user(tmp, (u8 __user *)arg);
		else
			retval = __get_user(tmp, (u32 __user *)arg);
		if (retval == 0) {
			u32	save = spi->mode;

			if (tmp & ~SPI_MODE_MASK) {
				retval = -EINVAL;
				break;
			}

			tmp |= spi->mode & ~SPI_MODE_MASK;
			spi->mode = (u16)tmp;
			retval = spi_setup(spi);
			if (retval < 0)
				spi->mode = save;
			else
				dev_dbg(&spi->dev, "spi mode %x\n", tmp);
		}
		break;
	case SPI_IOC_WR_LSB_FIRST:		
		retval = __get_user(tmp, (__u8 __user *)arg);
		if (retval == 0) {
			u32	save = spi->mode;

			if (tmp)
				spi->mode |= SPI_LSB_FIRST;
			else
				spi->mode &= ~SPI_LSB_FIRST;
			retval = spi_setup(spi);
			if (retval < 0)
				spi->mode = save;
			else
				dev_dbg(&spi->dev, "%csb first\n",
						tmp ? 'l' : 'm');
		}
		break;
	case SPI_IOC_WR_BITS_PER_WORD:	
		retval = __get_user(tmp, (__u8 __user *)arg);
		if (retval == 0) {
			u8	save = spi->bits_per_word;

			spi->bits_per_word = tmp;
			retval = spi_setup(spi);
			if (retval < 0)
				spi->bits_per_word = save;
			else
				dev_dbg(&spi->dev, "%d bits per word\n", tmp);
		}
		break;
	case SPI_IOC_WR_MAX_SPEED_HZ:	
		retval = __get_user(tmp, (__u32 __user *)arg);
		if (retval == 0) {
			u32	save = spi->max_speed_hz;

			spi->max_speed_hz = tmp;
			retval = spi_setup(spi);
			if (retval < 0)
				spi->max_speed_hz = save;
			else
				dev_dbg(&spi->dev, "%d Hz (max)\n", tmp);
		}
		break;	
	case 0x4A806B07:	//SPI_IOC_DISPLAY	
		copy_from_user(spidev->tx_buffer, (void __user *)arg, _IOC_SIZE(cmd));
		write_command(spidev->spi, SetColumnAddr);
		write_data(spidev->spi, 0x00);
		write_command(spidev->spi, SetPageAddr);	
		write_data(spidev->spi, 0x00);	
		write_command(spidev->spi, 0x01);
		retval =write_data_buffer(spidev->spi,spidev->tx_buffer,_IOC_SIZE(cmd));
		break;	
	case 0x40006B11:	//SPI_IOC_CONTRAST
	case 0x40006B12:	
	case 0x40006B13:	
	case 0x40006B14:	
	case 0x40006B15:	
	case 0x40006B16:	
	case 0x40006B17:	
	case 0x40006B18:
	case 0x40006B19:
	case 0x40006B1A:
		ContrastValue= (cmd & 0x0000000F)*4-2;
		write_command(spi, SetPMCtrl);		
		write_data(spi, 0x00);
		write_data(spi, (unsigned char)ContrastValue);	

		write_command(spi, 0x26);

		printk("[spidev_ioctl] cmd(%4x), _IOC_TYPE(%d),_IOC_DIR(%d), _IOC_SIZE(%d), ContrastValue(%d)\n", cmd, _IOC_TYPE(cmd), _IOC_DIR(cmd), _IOC_SIZE(cmd), ContrastValue);
		break;
	default:		
		/* segmented and/or full-duplex I/O request */
		if (_IOC_NR(cmd) != _IOC_NR(SPI_IOC_MESSAGE(0))
				|| _IOC_DIR(cmd) != _IOC_WRITE) {
			retval = -ENOTTY;
			break;
		}

		tmp = _IOC_SIZE(cmd);
		if ((tmp % sizeof(struct spi_ioc_transfer)) != 0) {
			retval = -EINVAL;
			break;
		}
		n_ioc = tmp / sizeof(struct spi_ioc_transfer);
		if (n_ioc == 0)
			break;

		/* copy into scratch area */
		ioc = kmalloc(tmp, GFP_KERNEL);
		if (!ioc) {
			retval = -ENOMEM;
			break;
		}
		if (__copy_from_user(ioc, (void __user *)arg, tmp)) {
			kfree(ioc);
			retval = -EFAULT;
			break;
		}
		printk("[spidev_ioctl] default(%4x), size(%d), n_ioc(%d) > spidev_message()\n", cmd, tmp, n_ioc);
		/* translate to spi_message, execute */
		retval = spidev_message(spidev, ioc, n_ioc);
		kfree(ioc);
		
		break;
	}

	mutex_unlock(&spidev->buf_lock);
	spi_dev_put(spi);
	return retval;
}

#ifdef CONFIG_COMPAT
static long
spidev_compat_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	return spidev_ioctl(filp, cmd, (unsigned long)compat_ptr(arg));
}
#else
#define spidev_compat_ioctl NULL
#endif /* CONFIG_COMPAT */

static int spidev_lcd_open(struct inode *inode, struct file *filp)
{
	struct spidev_data	*spidev;
	int			status = -ENXIO;
	mutex_lock(&device_list_lock);

	list_for_each_entry(spidev, &device_list, device_entry) {
		if (spidev->devt == inode->i_rdev) {
			status = 0;
			break;
		}
		
	}
	if (status) {
		pr_debug("spidev: nothing for minor %d\n", iminor(inode));
		goto err_find_dev;
	}
	if (!spidev->tx_buffer) {
		spidev->tx_buffer = kmalloc(FRAMEBUFF_SIZE, GFP_KERNEL);
		if (!spidev->tx_buffer) {
				dev_dbg(&spidev->spi->dev, "open/ENOMEM\n");
				status = -ENOMEM;
			goto err_find_dev;
			}
		}
	if (!spidev->rx_buffer) {
		spidev->rx_buffer = kmalloc(bufsiz, GFP_KERNEL);
		if (!spidev->rx_buffer) {
			dev_dbg(&spidev->spi->dev, "open/ENOMEM\n");
			status = -ENOMEM;
			goto err_alloc_rx_buf;
		}
	}

	spidev->users++;
	filp->private_data = spidev;
	nonseekable_open(inode, filp);
	
	dev_dbg(&spidev->spi->dev, "[JHL-20200615] spidev_lcd_open(OK)\n");
	mutex_unlock(&device_list_lock);
	return 0;

err_alloc_rx_buf:
	kfree(spidev->tx_buffer);
	spidev->tx_buffer = NULL;
err_find_dev:
	mutex_unlock(&device_list_lock);
	return status;
}

static int spidev_release(struct inode *inode, struct file *filp)
{
	struct spidev_data	*spidev;
	int			status = 0;

	mutex_lock(&device_list_lock);
	spidev = filp->private_data;
	filp->private_data = NULL;
	/* last close? */
	spidev->users--;
	if (!spidev->users) {
		int		dofree;

		kfree(spidev->tx_buffer);
		spidev->tx_buffer = NULL;

		kfree(spidev->rx_buffer);
		spidev->rx_buffer = NULL;

		/* ... after we unbound from the underlying device? */
		spin_lock_irq(&spidev->spi_lock);
		dofree = (spidev->spi == NULL);
		spin_unlock_irq(&spidev->spi_lock);

		if (dofree)
			kfree(spidev);
	}
	mutex_unlock(&device_list_lock);

	return status;
}

static const struct file_operations spidev_lcd_fops = {
	.owner =	THIS_MODULE,
	/* REVISIT switch to aio primitives, so that userspace
	 * gets more complete API coverage.  It'll simplify things
	 * too, except for the locking.
	 */
	.write =	spidev_lcd_write,
	.read =		spidev_lcd_read,
	.unlocked_ioctl = spidev_ioctl,
	.compat_ioctl = spidev_compat_ioctl,
	.open =		spidev_lcd_open,
	.release =	spidev_release,
	//.llseek =	no_llseek,
};

void write_command(struct spi_device *spi, unsigned char cmd)
{
    gpio_set_value(GPIO_RS, 0); 	// dc , command:0 
    spi_write(spi, &cmd,1);
}

void write_data(struct spi_device *spi,unsigned char data )
{
    // dc , data:1 
    gpio_set_value(GPIO_RS, 1); 
    spi_write(spi,&data,1);
}

void write_image(struct spi_device *spi,unsigned char data )
{ 	
    spi_write(spi,&data,1);
}

void read_data(struct spi_device *spi,unsigned char *data,char len )
{
    // dc , data:1
    gpio_set_value(GPIO_RS, 1); 
    spi_read(spi,data,len);
}


int write_data_buffer(struct spi_device *spi,unsigned char *data ,int len)
{
    gpio_set_value(GPIO_RS, 1); // dc , data:1 
    return spidev_sync_write(spi->dev.driver_data,data,len);
	
}

void addr_set(struct spi_device *spi, unsigned int x1,unsigned int y1,unsigned int x2,unsigned int y2)
{
        write_command(spi, 0x2a); 
        write_data(spi, x1>>8);
        write_data(spi, x1&0xff);
		write_data(spi, x2>>8);
        write_data(spi, x2&0xff);

        write_command(spi, 0x2b); 
        write_data(spi, y1>>8);
        write_data(spi, y1&0xff);
		write_data(spi, y2>>8);
        write_data(spi, y2&0xff);

        write_command(spi, 0x2c);
}

void clear_screen(struct spi_device *spi,unsigned short color)
{
	unsigned short i,j;
	unsigned char vh,vl;
	vh=color>>8;
	vl=color;
	addr_set(spi,0,0,LCD_W-1,LCD_H-1);
	for(i=0;i<LCD_H;i++)
	{
		for(j=0;j<LCD_W;j++)
		{
			write_data(spi,vh);
			write_data(spi,vl);
		}
	}
}

void clear_screen_buffer(struct spi_device *spi,unsigned short color)
{
	struct spidev_data	*spidev_lcd;
	unsigned short i,j;
	unsigned char *buf;
	char init=0;
	unsigned char vh,vl;
	
	spidev_lcd=(struct spidev_data	*)spi->dev.driver_data;
	if(spidev_lcd!=0)
		buf=spidev_lcd->tx_buffer;
	else
	{
		init=1;
		buf=kmalloc(FRAMEBUFF_SIZE, GFP_KERNEL);
	}

	vh=color>>8;
	vl=color;
	addr_set(spi,0,0,LCD_W-1,LCD_H-1);
	for(i=0;i<=LCD_H;i++)
	{
		for(j=0;j<=LCD_W;j++)
		{
			buf[(i*LCD_W+j)*2]=vh;
			buf[(i*LCD_W+j)*2+1]=vl;
		}
	}
	write_data_buffer(spi,buf,FRAMEBUFF_SIZE);
	if(init==1)
	{
		kfree(buf);
		init=0;
	}
}

#ifdef lcd_ili9341
void lcd_gpio_init(void)
{
	printk(KERN_WARNING"gpio_requst:%d\n",gpio_request(GPIO_RST,"spidev_lcd"));
	printk(KERN_WARNING"gpio_requst:%d\n",gpio_request(GPIO_RS,"spidev_lcd"));
    	gpio_direction_output(GPIO_RST, 0);
	gpio_direction_output(GPIO_RS, 0);
    	mdelay(50);
    	gpio_set_value(GPIO_RST, 1);
	mdelay(50);
}

void spi_lcd_init(struct spi_device *spi)
{
	printk("spi lcd init\n");
	printk("spi command end(No Init)\n");	
}

#elif defined lcd_st7789
spi_lcd_cmd cmds[]={
		{0x36,1,0},
		{0x3a,1,0},
		{0xb2,5,0},
		{0xb7,1,0},
		{0xbb,1,0},
		{0xc0,1,0},
		{0xc2,1,0},
		{0xc3,1,0},
		{0xc4,1,0},
		{0xc6,1,0},
		{0xca,1,0},
		{0xc8,1,0},
		{0x55,1,0},
		{0xd0,2,0},
		{0xe0,14,0},
		{0xe1,14,0},
		{0x29,0,0},
};
unsigned char spi_lcd_datas[] = {
	0x00,
	0x05,
	0x0c,0x0c,0x00,0x33,0x33,
	0x35,			//0xb7
	0x35,			//0xbb
	0x2c,			//0xc0
	0x01,			//0xc2
	0x13,			//0xc3
	0x20,			//0xc4
	0x0f,				//0xc6
	0x0f,				//0xca
	0x08,			//0xc8
	0x90,			//0x55
	0xa4,0xa1,		//0xd0
	0xd0,0x00,0x06,0x09,0x0b,0x2a,0x3c,0x55,0x4b,0x08,0x16,0x14,0x19,0x20,		//0xe0
	0xd0,0x00,0x06,0x09,0x0b,0x29,0x36,0x54,0x4b,0x0d,0x16,0x14,0x21,0x20,		//0xe1
};

void lcd_gpio_init(void)
{
	unsigned char buf[4]={0};
//	printk(KERN_WARNING"gpio_requst:%d\n",gpio_request(GPIO_RST,"spidev_lcd"));
//	printk(KERN_WARNING"gpio_requst:%d\n",gpio_request(GPIO_RS,"spidev_lcd"));
	gpio_direction_output(GPIO_RST, 0);
	gpio_direction_output(GPIO_RS, 0);
	//gpio_set_value(GPIO_RST, 1);
	gpio_direction_output(GPIO_RST, 1);
	mdelay(10);
	//gpio_set_value(GPIO_RST, 0);
	gpio_direction_output(GPIO_RST, 0);
	mdelay(100);
	//gpio_set_value(GPIO_RST, 1);
	gpio_direction_output(GPIO_RST, 1);
	mdelay(500);
	write_command(spi, 0x04);
	read_data(spi,buf,4);
	for(i=0;i<4;i++)
		printk("bufx[%d]:%p\n",i,buf[i]);
	write_command(spi, 0x01);
	mdelay(100);
	write_command(spi, 0x11);
	mdelay(120);
	
}

#elif defined lcd_ili9163
spi_lcd_cmd cmds[]={
		{0x11, 0, 120},
		{0xB1, 3, 0},
		{0xB2, 3, 0},
		{0xB3, 6, 0},
		{0xB4, 1, 0},
		{0xC0, 3, 0},
		{0xC1, 1, 0},
		{0xC2, 2, 0},
		{0xC3, 2, 0},
		{0xC4, 2, 0},
		{0xC5, 1, 0},
		{0x36, 1, 0},
		{0xE0, 16, 0},
		{0xE1, 16, 0},
		{0x3A, 1,  0},
		{0x29, 0, 0},
};

unsigned char spi_lcd_datas[] = {
    0x05, 0x3C,  0x3C,                     // command: 0xB1
    0x05, 0x3C,  0x3C,		           // command: 0xB2
    0x05, 0x3C,  0x3C, 05, 0x3C,  0x3C,    // command: 0xB3
    0x03,				// command: 0xB4
    0x28,  0x08, 0x04,                                // command: 0xC0
    0xC0,                                   // command: 0xC1
    0x0D, 0x00,                             // command: 0xC2
    0x8D,   0x2A,                                // command: 0xC3
    0x8D,  0xEE,                                  // command: 0xC4
    0x1A,                                   // command: 0xC5
    0xA0,                           // command: 0x36
    0x04, 0x22, 0x07, 0x0A, 0x2E, 0x30, 0x25, 0x2A, 0x28, 0x26, 0x2E, 0x3A, 0x00, 0x01, 0x03,0x13, //command: 0xE0
    0x04, 0x16, 0x06, 0x0D, 0x2D, 0x26, 0x23, 0x27, 0x27, 0x25, 0x2D, 0x3B, 0x00, 0x01, 0x04,0x13, //command: 0xE1
    0x05,   //command:0x3A
};

void  lcd_gpio_init(void)
{
//	printk(KERN_WARNING"mhwld[%s][%s:%d]\n",__func__,__FILE__,__LINE__);
//	printk(KERN_WARNING"gpio_requst:%d\n",gpio_request(GPIO_RST,"spidev_lcd"));
//	printk(KERN_WARNING"gpio_requst:%d\n",gpio_request(GPIO_RS,"spidev_lcd"));
//	printk(KERN_WARNING"gpio_requst:%d\n",gpio_request(GPIO_BL,"spidev_lcd"));
	gpio_direction_output(GPIO_BL, 1);//backlight turn on
    	gpio_direction_output(GPIO_RST, 0);
	gpio_direction_output(GPIO_RS, 0);
    	mdelay(50);
    	gpio_set_value(GPIO_RST, 1);
	mdelay(50);
}
#endif


/*-------------------------------------------------------------------------*/

/* The main reason to have this class is to make mdev/udev create the
 * /dev/spidevB.C character device nodes exposing our userspace API.
 * It also simplifies memory management.
 */

static struct class *spidev_lcd_class;

/*-------------------------------------------------------------------------*/

static int spidev_lcd_probe(struct spi_device *spi)
{
	struct spidev_data	*spidev_lcd;
	int			status;
	unsigned long		minor;

	/* Allocate driver data */
	spidev_lcd = kzalloc(sizeof(*spidev_lcd), GFP_KERNEL);
	if (!spidev_lcd)
		return -ENOMEM;
	if (!spidev_lcd->tx_buffer) {
		spidev_lcd->tx_buffer = kmalloc(FRAMEBUFF_SIZE, GFP_KERNEL);
		if (!spidev_lcd->tx_buffer) {
				dev_dbg(&spidev_lcd->spi->dev, "open/ENOMEM\n");
				status = -ENOMEM;
			goto err_find_dev;
			}
		}
	if (!spidev_lcd->rx_buffer) {
		spidev_lcd->rx_buffer = kmalloc(bufsiz, GFP_KERNEL);
		if (!spidev_lcd->rx_buffer) {
			dev_dbg(&spidev_lcd->spi->dev, "open/ENOMEM\n");
			status = -ENOMEM;
			goto err_alloc_rx_buf;
		}
	}
	/* Initialize the driver data */
	spidev_lcd->spi = spi;
	spin_lock_init(&spidev_lcd->spi_lock);
	mutex_init(&spidev_lcd->buf_lock);

	INIT_LIST_HEAD(&spidev_lcd->device_entry);

	/* If we can allocate a minor number, hook up this device.
	 * Reusing minors is fine so long as udev or mdev is working.
	 */
	mutex_lock(&device_list_lock);
	minor = find_first_zero_bit(minors, N_SPI_MINORS);
	if (minor < N_SPI_MINORS) {
		struct device *dev;

		spidev_lcd->devt = MKDEV(SPIDEV_MAJOR, minor);
		dev = device_create(spidev_lcd_class, &spi->dev, spidev_lcd->devt,
				    spidev_lcd, "spidev%d.%d",
				    spi->master->bus_num, spi->chip_select);
		status = PTR_ERR_OR_ZERO(dev);
	} else {
		dev_dbg(&spi->dev, "no minor number available!\n");
		status = -ENODEV;
	}
	if (status == 0) {
		set_bit(minor, minors);
		list_add(&spidev_lcd->device_entry, &device_list);
	}
	mutex_unlock(&device_list_lock);
	if (status == 0)
		spi_set_drvdata(spi, spidev_lcd);
	else
		kfree(spidev_lcd);
	spi->max_speed_hz = 4800000;
	spi->mode &=~(0xf);
	spi->mode |= SPI_MODE_0;
	spi_setup(spi);
	spi_lcd_init(spi);
return status;
	
err_alloc_rx_buf:
	kfree(spidev_lcd->tx_buffer);
	spidev_lcd->tx_buffer = NULL;
err_find_dev:
	mutex_unlock(&device_list_lock);


}

static int spidev_lcd_remove(struct spi_device *spi)
{
	struct spidev_data	*spidev = spi_get_drvdata(spi);

	/* make sure ops on existing fds can abort cleanly */
	spin_lock_irq(&spidev->spi_lock);
	spidev->spi = NULL;
	spin_unlock_irq(&spidev->spi_lock);
	gpio_free(GPIO_RST);
	gpio_free(GPIO_RS);
	/* prevent new opens */
	mutex_lock(&device_list_lock);
	list_del(&spidev->device_entry);
	device_destroy(spidev_lcd_class, spidev->devt);
	clear_bit(MINOR(spidev->devt), minors);
	kfree(spidev->tx_buffer);
	kfree(spidev->rx_buffer);
	if (spidev->users == 0)
		kfree(spidev);
	mutex_unlock(&device_list_lock);

	return 0;
}

static const struct of_device_id spidev_lcd_dt_ids[] = {
	{ .compatible = "ILI9341", },
	{ .compatible = "ILI9163", },
	{ .compatible = "ST7789V", },
	{},
};

MODULE_DEVICE_TABLE(of, spidev_lcd_dt_ids);

static struct spi_driver spidev_lcd_spi_driver = {
	.driver = {
		.name =		"spidev_lcd",
		.owner =	THIS_MODULE,
		.of_match_table = of_match_ptr(spidev_lcd_dt_ids),
	},
	.probe =	spidev_lcd_probe,
	.remove =	spidev_lcd_remove,

	/* NOTE:  suspend/resume methods are not necessary here.
	 * We don't do anything except pass the requests to/from
	 * the underlying controller.  The refrigerator handles
	 * most issues; the controller driver handles the rest.
	 */
};

/*-------------------------------------------------------------------------*/

static int __init spidev_lcd_init(void)
{
	int status;
	/* Claim our 256 reserved device numbers.  Then register a class
	 * that will key udev/mdev to add/remove /dev nodes.  Last, register
	 * the driver which manages those device numbers.
	 */
	BUILD_BUG_ON(N_SPI_MINORS > 256);
	status = register_chrdev(SPIDEV_MAJOR, "spi_lcd", &spidev_lcd_fops);
	if (status < 0)
		return status;
	
	spidev_lcd_class = class_create(THIS_MODULE, "spidev_lcd");
	if (IS_ERR(spidev_lcd_class)) {
		status = PTR_ERR(spidev_lcd_class);
		goto error_class;
	}

	status = spi_register_driver(&spidev_lcd_spi_driver);
	if (status < 0)
		goto error_register;

	if (busnum != -1 && chipselect != -1) {
		struct spi_board_info chip = {
					.modalias	= "spidev_lcd",
					.mode		= spimode,
					.bus_num	= busnum,
					.chip_select	= chipselect,
					.max_speed_hz	= maxspeed,
		};
		struct spi_master *master;
		master = spi_busnum_to_master(busnum);
		if (!master) {
			status = -ENODEV;
			goto error_busnum;
		}
		spi = spi_new_device(master, &chip);
		if (!spi) {
			status = -EBUSY;
			goto error_mem;
		}
		dev_dbg(&spi->dev, "busnum=%d cs=%d bufsiz=%d maxspeed=%d",
			busnum, chipselect, bufsiz, maxspeed);
	}
	return 0;
error_mem:
error_busnum:
	spi_unregister_driver(&spidev_lcd_spi_driver);
error_register:
	class_destroy(spidev_lcd_class);
error_class:
	unregister_chrdev(SPIDEV_MAJOR, spidev_lcd_spi_driver.driver.name);
	return status;
}
module_init(spidev_lcd_init);

static void __exit spidev_lcd_exit(void)
{
	if (spi) {
		spi_unregister_device(spi);
		spi = NULL;
	}
	spi_unregister_driver(&spidev_lcd_spi_driver);
	class_destroy(spidev_lcd_class);
	unregister_chrdev(SPIDEV_MAJOR, spidev_lcd_spi_driver.driver.name);
}
module_exit(spidev_lcd_exit);

MODULE_AUTHOR("Andrea Paterniani, <a.paterniani@swapp-eng.it>");
MODULE_DESCRIPTION("User mode SPI device interface");
MODULE_LICENSE("GPL");
MODULE_ALIAS("spi:spidev_lcd");


